#include "config.h"
#include <ctype.h>
#include "canonicalize.h"
#include "mystring.h"

struct result
{
  const char* next;
  mystring str;
  mystring comment;
  mystring list;

  result(const result&);
  result(const char* = 0);
  result(const char*, const mystring&, const mystring&, const mystring&);
  bool operator!() const
    {
      return !next;
    }
  operator bool() const
    {
      return next;
    }
};

result::result(const char* n)
  : next(n)
{
}

result::result(const char* n, const mystring& s,
	       const mystring& c, const mystring& l)
  : next(n), str(s), comment(c), list(l)
{
}

result::result(const result& r)
  : next(r.next), str(r.str), comment(r.comment), list(r.list)
{
}

#ifndef TRACE
#define ENTER(R)
#define FAIL(MSG) return result()
#define RETURNR(R) return R
#define RETURN(N,S,C,L) return result(N,S,C,L);
#else
#include "fdbuf.h"
static const char indentstr[] = "                       ";
static const char* indent = indentstr + sizeof indentstr - 1;
#define ENTER(R) do{ fout.write(indent--); fout.write(__FUNCTION__); fout.write(": "); fout.write(ptr); fout.write(": "); fout.writeln(R); }while(0)
#define FAIL(MSG) do{ fout.write(++indent); fout.write(__FUNCTION__); fout.write(": failed: "); fout.writeln(MSG); return result(); }while(0)
#define RETURNR(R) do{ fout.write(++indent); fout.write(__FUNCTION__); fout.write(": succeded str="); fout.write(R.str); fout.write(" comment="); fout.writeln(R.comment); return (R); }while(0)
#define RETURN(N,S,C,L) do{ result r(N,S,C,L); RETURNR(r); }while(0)
#endif

#define RULE(X) static result match_##X(const char* ptr)
#define SKIPSPACE do{ while(*ptr && isspace(*ptr)) ++ptr; }while(0)
#define SKIPDELIM(N) do{ ptr = skipdelim(ptr, N); }while(0)
#define MATCHCHAR(X) do{ if(*ptr != X) FAIL("*ptr is not " #X); else ++ptr; }while(0)
#define OR_RULE(ALT1,ALT2) { result r=match_##ALT1(ptr); if(r) RETURNR(r); }{ result r=match_##ALT2(ptr); if(r) RETURNR(r); } FAIL("did not match " #ALT1 " OR " #ALT2);

#define COLON ':'
#define SEMICOLON ';'
#define LABRACKET '<'
#define RABRACKET '>'
#define AT '@'
#define ESCAPE '\\'
#define PERIOD '.'
#define LSQBRACKET '['
#define RSQBRACKET ']'
#define QUOTE '"'
#define CR '\n'
#define LPAREN '('
#define RPAREN ')'
#define COMMA ','

// quoted-pair = ESCAPE CHAR
static bool isqpair(const char* ptr)
{
  return *ptr && *ptr == ESCAPE &&
    *(ptr+1);
}

static bool isctext(char c)
{
  return c && c != LPAREN && c != RPAREN &&
    c != ESCAPE && c != CR;
}

RULE(comment)
{
  ENTER("LPAREN *(ctext / quoted-pair / comment) RPAREN");
  SKIPSPACE;
  const char* start = ptr;
  MATCHCHAR(LPAREN);
  for(;;) {
    if(isctext(*ptr))
      ++ptr;
    else if(isqpair(ptr))
      ptr += 2;
    else {
      result r = match_comment(ptr);
      if(r)
	ptr = r.next;
      else
	break;
    }
  }
  MATCHCHAR(RPAREN);
  RETURN(ptr, mystring(" ", mystringtmp(start, ptr-start)), "", "");
}

static const char* skipdelim(const char* ptr, mystring& comment)
{
  for(;;) {
    while(*ptr && isspace(*ptr))
      ++ptr;
    if(*ptr != LPAREN)
      break;
    result r = match_comment(ptr);
    if(r) {
      ptr = r.next;
      comment += r.str;
    }
    else
      break;
  }
  return ptr;
}

void escape_addr(mystring&)
{
}

static bool isspecial(char c)
{
  switch(c) {
  case LPAREN: case RPAREN:
  case LABRACKET: case RABRACKET:
  case LSQBRACKET: case RSQBRACKET:
  case AT: case COMMA:
  case SEMICOLON: case COLON:
  case ESCAPE: case QUOTE: case PERIOD:
    return true;
  default:
    return false;
  }
}

static bool isctl(char c)
{
  return (c >= 0 && c <= 31) || (c == 127);
}
  
RULE(atom)
{
  ENTER("1*<any CHAR except specials, SPACE and CTLs>");
  mystring comment;
  SKIPDELIM(comment);
  const char* start = ptr;
  while(*ptr &&
	!isspace(*ptr) &&
	!isspecial(*ptr) &&
	!isctl(*ptr))
    ++ptr;
  if(start == ptr)
    FAIL("no CHARs matched");
  RETURN(ptr, mystring(start, ptr-start), comment, "");
}

static bool isqtext(char c)
{
  return c && c != QUOTE && c != ESCAPE && c != CR;
}

RULE(quoted_string)
{
  ENTER("QUOTE *(qtext/quoted-pair) QUOTE");
  mystring comment;
  SKIPDELIM(comment);
  MATCHCHAR(QUOTE);
  const char* start = ptr;
  for(; *ptr; ++ptr) {
    if(isqtext(*ptr))
       continue;
    else if(isqpair(ptr))
      ++ptr;
    else
      break;
  }
  MATCHCHAR(QUOTE);
  RETURN(ptr, mystring(start, ptr-start-1), comment, "");
}
  
static bool isdtext(char c)
{
  return c && c != LSQBRACKET && c != RSQBRACKET &&
    c != ESCAPE && c != CR;
}

RULE(domain_literal)
{
  ENTER("LSQBRACKET *(dtext/quoted-pair) RSQBRACKET");
  SKIPSPACE;
  MATCHCHAR(LSQBRACKET);
  SKIPSPACE;
  const char* start = ptr;
  mystring str;
  for(; *ptr; ++ptr) {
    if(isdtext(*ptr))
      continue;
    else if(isqpair(ptr))
      ++ptr;
    else
      break;
  }
  SKIPSPACE;
  MATCHCHAR(RSQBRACKET);
  RETURN(ptr, mystring(start, ptr-start), "", "");
}

RULE(sub_domain)
{
  // NOTE: domain-ref  =  atom
  ENTER("domain-ref / domain-literal");
  OR_RULE(atom, domain_literal);
}

RULE(domain)
{
  ENTER("sub-domain *(PERIOD sub-domain)");
  result r = match_sub_domain(ptr);
  if(!r) FAIL("did not match sub-domain");
  mystring comment;
  for(;;) {
    ptr = r.next;
    SKIPDELIM(comment);
    r.next = ptr;
    if(*ptr++ != PERIOD)
      break;
    result r1 = match_sub_domain(ptr);
    if(!r1) break;
    r.next = r1.next;
    r.str += PERIOD;
    r.str += r1.str;
    comment += r1.comment;
  }
  r.comment += comment;
  RETURNR(r);
}

RULE(route)
{
  ENTER("1#(AT domain) COLON");
  unsigned count=0;
  mystring str;
  mystring comment;
  for(;;) {
    if(*ptr != AT) break;
    ++ptr;
    result r = match_domain(ptr);
    if(!r) FAIL("did not match domain");
    str += mystring("@", r.str);
    comment += r.comment;
    ++count;
    ptr = r.next;
  }
  if(count == 0)
    FAIL("matched no domains");
  SKIPDELIM(comment);
  MATCHCHAR(COLON);
  RETURN(ptr, str, comment, "");
}

RULE(word)
{
  ENTER("atom / quoted-string");
  OR_RULE(atom, quoted_string);
}

RULE(local_part)
{
  ENTER("word *(PERIOD word)");
  result r = match_word(ptr);
  if(!r) FAIL("did not match word");
  for(;;) {
    ptr = r.next;
    SKIPDELIM(r.comment);
    r.next = ptr;
    if(*ptr++ != PERIOD)
      break;
    result r1 = match_word(ptr);
    if(!r1)
      break;
    r.str += PERIOD;
    r.str += r1.str;
    r.comment += r1.comment;
    r.next = r1.next;
  }
  RETURNR(r);
}
  
RULE(addr_spec)
{
  ENTER("local-part *( AT domain )");
  result r = match_local_part(ptr);
  if(!r) FAIL("did not match local-part");
  mystring domain;
  for(;;) {
    ptr = r.next;
    SKIPDELIM(r.comment);
    r.next = ptr;
    if(*ptr++ != AT)
      break;
    result r2 = match_domain(ptr);
    if(!r2) break;
    if(!!domain)
      r.str = mystring(r.str, "@", domain);
    domain = r2.str;
    r.comment += r2.comment;
    r.next = r2.next;
  }
  canonicalize(domain);
  RETURN(r.next, mystring(r.str, "@", domain), r.comment,
	 mystring(r.str, "@", domain, "\n"));
}

RULE(route_addr) 
{
  ENTER("LABRACKET [route] addr-spec RABRACKET");
  mystring comment;
  SKIPDELIM(comment);
  MATCHCHAR(LABRACKET);
  result r1 = match_route(ptr);
  if(r1) ptr = r1.next;
  comment += r1.comment;
  result r2 = match_addr_spec(ptr);
  if(!r2) FAIL("did not match addr-spec");
  ptr = r2.next;
  comment += r2.comment;
  SKIPDELIM(comment);
  MATCHCHAR(RABRACKET);
  RETURN(ptr, mystring("<", r2.str, ">", comment), "",
	 mystring(r2.str, "\n"));
}

RULE(phrase)
{
  ENTER("word *word");
  result r1 = match_word(ptr);
  if(!r1) FAIL("did not match word");
  for(;;) {
    result r2 = match_word(r1.next);
    if(!r2)
      break;
    r1.str = mystring(r1.str, " ", r2.str);
    r1.comment += r2.comment;
    r1.next = r2.next;
  }
  RETURNR(r1);
}

static bool isrtext(char c)
{
  return c && c != LABRACKET && c != RABRACKET &&
    c != ESCAPE && c != CR;
}

RULE(route_phrase)
{
  ENTER("*(rtext/quoted-pair)");
  SKIPSPACE;
  const char* start = ptr;
  for(;; ++ptr) {
    if(isrtext(*ptr))
      continue;
    else if(isqpair(ptr))
      ++ptr;
    else
      break;
  }
  RETURN(ptr, mystring(start, ptr-start), "", "");
}

RULE(route_spec)
{
  ENTER("route-phrase route-addr");
  result r1 = match_route_phrase(ptr);
  if(!r1) FAIL("did not match route-phrase");
  result r2 = match_route_addr(r1.next);
  if(!r2) FAIL("did not match route-addr");
  r2.str = mystring(r1.str, r1.comment, " ", r2.str, r2.comment);
  RETURNR(r2);
}

RULE(mailbox)
{
  ENTER("route-spec / addr-spec");
  OR_RULE(route_spec, addr_spec);
}

RULE(mailboxes)
{
  ENTER("mailbox *(*(COMMA) mailbox)");
  result r1 = match_mailbox(ptr);
  if(!r1) FAIL("did not match mailbox");
  r1.str += r1.comment;
  r1.comment = "";
  for(;;) {
    ptr = r1.next;
    for(;;) {
      SKIPDELIM(r1.str);
      if(*ptr == COMMA) ++ptr;
      else break;
    }
    if(!*ptr)
      break;
    result r2 = match_mailbox(ptr);
    if(!r2) break;
    r1.next = r2.next;
    r1.str = mystring(r1.str, ", ", r2.str, r2.comment);
    r1.list += r2.list;
  }
  SKIPDELIM(r1.str);
  r1.next = ptr;
  RETURNR(r1);
}

RULE(group)
{
  ENTER("phrase COLON [#mailboxes] SEMICOLON");
  result r1 = match_phrase(ptr);
  if(!r1) FAIL("did not match phrase");
  ptr = r1.next;
  SKIPSPACE;
  MATCHCHAR(COLON);
  result r2 = match_mailboxes(ptr);
  if(r2) ptr = r2.next;
  mystring comment;
  SKIPDELIM(comment);
  MATCHCHAR(SEMICOLON);
  RETURN(ptr, mystring(r1.str, ": ", r2.str, r2.comment, comment, ";"), "", r2.list);
}

RULE(address)
{
  ENTER("group / mailbox");
  OR_RULE(group, mailbox);
}

RULE(part)
{
  ENTER("address / comment");
  OR_RULE(address, comment);
}

RULE(addresses)
{
  ENTER("part *(*(COMMA) part) EOF");
  result r1 = match_part(ptr);
  if(!r1) FAIL("did not match part");
  r1.str += r1.comment;
  r1.comment = "";
  for(;;) {
    ptr = r1.next;
    for(;;) {
      SKIPDELIM(r1.str);
      if(*ptr == COMMA) ++ptr;
      else break;
    }
    if(!*ptr)
      break;
    result r2 = match_part(ptr);
    if(!r2) break;
    r1.next = r2.next;
    r1.str = mystring(r1.str, ", ", r2.str, r2.comment);
    r1.list += r2.list;
  }
  SKIPDELIM(r1.str);
  if(*ptr) FAIL("Rule ended before EOF");
  RETURNR(r1);
}
    
bool parse_addresses(mystring& line, mystring& list)
{
  result r = match_addresses(line.c_str());
  if(r) {
    line = r.str;
    list = r.list;
    return true;
  }
  else
    return false;
}
